/*;
	Audio Overload SDK - main driver.  for demonstration only, not user friendly!

	Copyright (c) 2007-2009 R. Belmont and Richard Bannister.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

	* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
	* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
	* Neither the names of R. Belmont and Richard Bannister nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
	"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
	LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
	A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
	NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ao.h"
#include "eng_protos.h"

/* file types */
static uint32 type;
static char *base_dir=0;
#define DEBUG_LOADER


/* ao_get_lib: called to load secondary files */
int ao_get_lib(char *filename, uint8 **buffer, uint64 *length)
{
	uint8 *filebuf;
	uint32 size;
	FILE *auxfile;
	char *lib_path;
	
	if(base_dir!=0){
		lib_path=malloc(strlen(base_dir)+strlen(filename)+5);
		sprintf(lib_path,"%s%s",base_dir,filename);
		auxfile = fopen(lib_path, "rb");
	}else{
		auxfile = fopen(filename, "rb");
	}

	if (!auxfile)
	{
		printf("Unable to find auxiliary file %s\n", filename);
		return AO_FAIL;
	}

	fseek(auxfile, 0, SEEK_END);
	size = ftell(auxfile);
	fseek(auxfile, 0, SEEK_SET);

	filebuf = malloc(size);

	if (!filebuf)
	{
		fclose(auxfile);
		printf("ERROR: could not allocate %d bytes of memory\n", size);
		return AO_FAIL;
	}

	fread(filebuf, size, 1, auxfile);
	fclose(auxfile);

	*buffer = filebuf;
	*length = (uint64)size;

	return AO_SUCCESS;
}

__stdcall int32 std_psf2_start(uint8 *file_buffer, uint32 length){	return psf2_start(file_buffer,length);	}
__stdcall int32 std_psf2_gen(int16 *buffer, uint32 samples)		{	return psf2_gen(buffer,samples);	}
__stdcall int32 std_psf2_stop(void)						{	return psf2_stop();				}
__stdcall int32 std_psf2_command(int32 cmd, int32 param)		{	return psf2_command(cmd,param);	}

__stdcall int ps2_test(char* filepath);

int main(int argv, char *argc[])
{
	if (argv < 2)
	{
		printf("Error: must specify a filename!\n");
		return -1;
	}	
	ps2_test(argc[1]);
	return 1;
}

// stub for MAME stuff
int change_pc(int foo){}


//set wave data infomation
__stdcall void get_audio_info(int *feq,unsigned char *ch,unsigned char *bits){
	*feq=44100;
	*ch=2;
	*bits=16;
}

//Cu(_lib)TtH_̎w(path must exist '\' char in before null)
__stdcall void set_lib_dir(char* dir_path){
	if(base_dir!=0)free(base_dir);		//base_dirɎgĂ
	base_dir=malloc(strlen(dir_path)+5);	//dir_paht̕m
	strcpy(base_dir,dir_path);			//dir_pathbase_dirɃRs[
}

/* redirect stubs to interface the Z80 core to the QSF engine */
uint8 memory_read(uint16 addr)
{
	return qsf_memory_read(addr);
}

uint8 memory_readop(uint16 addr)
{
	return memory_read(addr);
}

uint8 memory_readport(uint16 addr)
{
	return qsf_memory_readport(addr);
}

void memory_write(uint16 addr, uint8 byte)
{
	qsf_memory_write(addr, byte);
}

void memory_writeport(uint16 addr, uint8 byte)
{
	qsf_memory_writeport(addr, byte);
}

__stdcall int ps2_test(char* filepath)
{
	FILE *file;
	uint8 *buffer;
	uint32 size, filesig;

	printf("AOSDK test program v1.0 by R. Belmont [AOSDK release 1.4.8]\nCopyright (c) 2007-2009 R. Belmont and Richard Bannister - please read license.txt for license details\n\n");
	printf("init test , base_dir = %X",base_dir);
	// check if an argument was given

	file = fopen(filepath, "rb");

	if (!file)
	{
		printf("ERROR: could not open file %s\n", filepath);
		return -1;
	}

	// get the length of the file by seeking to the end then reading the current position
	fseek(file, 0, SEEK_END);
	size = ftell(file);
	// reset the pointer
	fseek(file, 0, SEEK_SET);
	printf("file size=%d\n",size);

	buffer = malloc(size);

	if (!buffer)
	{
		fclose(file);
		printf("ERROR: could not allocate %d bytes of memory\n", size);
		return -1;
	}

	// read the file
	fread(buffer, size, 1, file);
	fclose(file);

	

	if (std_psf2_start(buffer, size) != AO_SUCCESS)
	{
		free(buffer);
		printf("ERROR: Engine rejected file!\n");
		return -1;
	}else printf("Start OK\n");
	const nSample = 735;

	printf("\n\nPlaying.  Press CTRL-C to stop.\n");

	FILE *out;	int i,ret;	int16 *sb;
	sb=malloc(nSample*4*4);
	memset(sb,0,nSample*4*4);
	out=fopen("test_aodll.rwav","wb");
	if(out==0){
		printf("file output error!!!!\n");
		return 1;
	}
	printf("handle=%d\n",out);
	
	for(i=0;i<10000;i++){
		ret=std_psf2_gen(sb,nSample);		// format:44100Hz 2ch 16bit
		fwrite(sb,4,nSample,out);//16bit * 2ch * 1024
		//printf("%d ",ret);
		if(ret==AO_EOF){
			printf("[End Of Song]left:%d\n",10000-i);
			break;
		}else if(ret==AO_FAIL){
			printf("EMURATION ERROR!!! AO_FAIL!!!!!\n");
			break;
		}
		if(ao_song_done!=0) printf("ao_song_done!!\n");
	}
	printf("exit loop\n");
	fclose(out);
	free(buffer);

	return 1;
}